import argparse
import sys
import struct
import msgpack
import logging

# --------------------------------------------------------------------------------
# Logger setup with optional color formatting for different log levels
# --------------------------------------------------------------------------------

COLOR_RESET  = "\033[0m"
COLOR_RED    = "\033[31m"
COLOR_GREEN  = "\033[32m"
COLOR_YELLOW = "\033[33m"
COLOR_BLUE   = "\033[34m"

class ColorFormatter(logging.Formatter):
    """
    Simple color formatter using ANSI escape codes to highlight log levels.
    """
    LEVEL_COLORS = {
        logging.DEBUG: COLOR_BLUE,
        logging.INFO: COLOR_GREEN,
        logging.WARNING: COLOR_YELLOW,
        logging.ERROR: COLOR_RED,
        logging.CRITICAL: COLOR_RED
    }

    def format(self, record):
        level_color = self.LEVEL_COLORS.get(record.levelno, COLOR_RESET)
        msg = super().format(record)
        return f"{level_color}{msg}{COLOR_RESET}"

logger = logging.getLogger("cdae_dump_info")
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(ColorFormatter("[%(levelname)s] %(message)s"))
logger.addHandler(handler)

# --------------------------------------------------------------------------------
# Material flags decoding helper, from tsMaterialList.h
# --------------------------------------------------------------------------------

def decode_material_flags(flags: int) -> str:
    """
    Decodes the material flags from tsMaterialList.h:

        S_Wrap             = BNG_BIT(0),
        T_Wrap             = BNG_BIT(1),
        Translucent        = BNG_BIT(2),
        Additive           = BNG_BIT(3),
        Subtractive        = BNG_BIT(4),
        SelfIlluminating   = BNG_BIT(5),
        NeverEnvMap        = BNG_BIT(6),
        NoMipMap           = BNG_BIT(7),
        MipMap_ZeroBorder  = BNG_BIT(8),
        AuxiliaryMap       = BNG_BIT(27..31)
    """
    results = []
    if flags & 0x00000001:
        results.append("S_Wrap")
    if flags & 0x00000002:
        results.append("T_Wrap")
    if flags & 0x00000004:
        results.append("Translucent")
    if flags & 0x00000008:
        results.append("Additive")
    if flags & 0x00000010:
        results.append("Subtractive")
    if flags & 0x00000020:
        results.append("SelfIlluminating")
    if flags & 0x00000040:
        results.append("NeverEnvMap")
    if flags & 0x00000080:
        results.append("NoMipMap")
    if flags & 0x00000100:
        results.append("MipMap_ZeroBorder")
    # bits 27..31 => 0xF8000000
    if flags & 0xF8000000:
        results.append("AuxiliaryMap")
    return ", ".join(results) if results else "none"

# --------------------------------------------------------------------------------
# Main parsing function: parse_v30_file
# --------------------------------------------------------------------------------

def parse_v30_file(filepath: str):
    """
    Parses a v30 TSShape file in the exact order it was written by _write_v30(),
    with no partial skipping.

    References:
      - tsShape.cpp: TSShape::_write_v30
      - This script reads fields in the same sequence they are `packer.pack_*`ed.
    """

    # 1) Open the file and read the initial 4 bytes to check shape version.
    with open(filepath, "rb") as f:
        version_bytes = f.read(4)
        if len(version_bytes) < 4:
            logger.error("File is too small to contain version data.")
            return
        version_val = struct.unpack("<I", version_bytes)[0]
        exporter_version = version_val >> 16
        shape_version = version_val & 0xFF
        if shape_version != 30:
            logger.error(f"Not a v30 shape (found shape_version={shape_version}, exporter={exporter_version}). Stopping.")
            return

        # Read the remainder of the file into memory and feed it to msgpack
        bin_data = f.read()

    unpacker = msgpack.Unpacker()
    unpacker.feed(bin_data)

    # --------------------------------------------------------------------------------
    # Helpers to read basic data from the msgpack stream
    # --------------------------------------------------------------------------------

    def next_obj():
        """
        Reads the next object from the msgpack.Unpacker.
        Returns None if we run out of data.
        """
        try:
            return next(unpacker)
        except StopIteration:
            return None

    def read_int(label="(int)"):
        """
        Reads either an int or a float that is integral (which we'll convert to int).
        Exits with error if the value is not integral.
        """
        val = next_obj()
        if isinstance(val, int):
            return val
        if isinstance(val, float):
            if abs(val - round(val)) < 1e-9:
                return int(round(val))
            logger.error(f"read_int('{label}'): got float={val} but not integral. Exiting.")
            sys.exit(1)
        logger.error(f"read_int('{label}'): expected int, got {type(val)} => {val}. Exiting.")
        sys.exit(1)

    def read_uint32(label="(uint32)"):
        """
        Same as read_int but must be non-negative.
        """
        val = read_int(label)
        if val < 0:
            logger.error(f"read_uint32('{label}'): got negative {val}. Exiting.")
            sys.exit(1)
        return val

    def read_float(label="(float)"):
        """
        Reads a float or an int -> float, else errors.
        """
        val = next_obj()
        if isinstance(val, float):
            return val
        if isinstance(val, int):
            return float(val)
        logger.error(f"read_float('{label}'): expected float/int, got {type(val)} => {val}. Exiting.")
        sys.exit(1)

    def read_str(label="(string)"):
        """
        Reads a string from the stream.
        """
        val = next_obj()
        if not isinstance(val, str):
            logger.error(f"read_str('{label}'): expected string, got {type(val)} => {val}. Exiting.")
            sys.exit(1)
        return val

    def read_point3f(label="(Point3F)"):
        """
        Reads either a 3-element list [x, y, z] or a 12-byte buffer -> <3f
        """
        val = next_obj()
        if isinstance(val, list):
            if len(val) != 3:
                logger.error(f"read_point3f('{label}'): list length {len(val)} != 3. Exiting.")
                sys.exit(1)
            return [float(x) for x in val]
        elif isinstance(val, bytes):
            if len(val) != 12:
                logger.error(f"read_point3f('{label}'): need 12 bytes, got {len(val)}. Exiting.")
                sys.exit(1)
            return list(struct.unpack("<3f", val))
        logger.error(f"read_point3f('{label}'): got {type(val)}, not list/bytes. Exiting.")
        sys.exit(1)

    def read_box6f(label="(Box3F)"):
        """
        Reads either a 6-element list [minX, minY, minZ, maxX, maxY, maxZ]
        or a 24-byte buffer -> <6f
        """
        val = next_obj()
        if isinstance(val, list):
            if len(val) != 6:
                logger.error(f"read_box6f('{label}'): list length {len(val)} != 6. Exiting.")
                sys.exit(1)
            return [float(x) for x in val]
        elif isinstance(val, bytes):
            if len(val) != 24:
                logger.error(f"read_box6f('{label}'): need 24 bytes, got {len(val)}. Exiting.")
                sys.exit(1)
            return list(struct.unpack("<6f", val))
        logger.error(f"read_box6f('{label}'): got {type(val)}, not list/bytes. Exiting.")
        sys.exit(1)

    def read_bytes(label="(bytes)"):
        """
        Reads a bytes object. Exits if we don't get 'bytes'.
        """
        val = next_obj()
        if not isinstance(val, bytes):
            logger.error(f"read_bytes('{label}'): expected bytes, got {type(val)} => {val}. Exiting.")
            sys.exit(1)
        return val

    # --------------------------------------------------------------------------------
    # Vector reading logic: 'pack_vector(...)' in C++ => read_vector() here
    # We also keep a dictionary of "expected" element sizes for debugging.
    # --------------------------------------------------------------------------------

    vector_type_expectations = {
        "nodes":                      ("Node data?",               20),
        "objects":                    ("TSShape::Object",          24),
        "subShapeFirstNode":          ("int",                      4),
        "subShapeFirstObject":        ("int",                      4),
        "subShapeNumNodes":           ("int",                      4),
        "subShapeNumObjects":         ("int",                      4),
        "defaultRotations":           ("Quat16",                   8),
        "defaultTranslations":        ("Point3F",                  12),
        "nodeRotations":              ("Quat16",                   8),
        "nodeTranslations":           ("Point3F",                  12),
        "nodeUniformScales":          ("float",                    4),
        "nodeAlignedScales":          ("Point3F",                  12),
        "nodeArbitraryScaleFactors":  ("Point3F",                  12),
        "nodeArbitraryScaleRots":     ("Quat16",                   8),
        "groundTranslations":         ("Point3F",                  12),
        "groundRotations":            ("Quat16",                   8),
        "objectStates":               ("TSShape::ObjectState",     12),  # 1 float + 2 int
        "triggers":                   ("TSShape::Trigger",         8),   # 1 u32 + 1 float
        "details":                    ("TSShape::Detail",          52),  # 13 fields total
        "mesh.verts":                 ("Point3F?",                 12),
        "mesh.tverts":                ("Point2F",                  8),
        "mesh.tverts2":               ("Point2F",                  8),
        "mesh.colors":                ("RGBA",                     4),
        "mesh.norms":                 ("Point3F",                  12),
        "mesh.encodedNorms":          ("PackedNorm",               1),
        "mesh.primitives":            ("TSDrawPrimitive",          12),
        "mesh.indices":               ("uint32?",                  4),
        "mesh.tangents":              ("Tangent?",                 16),
        "(vector)":                   ("Unknown",                  None),
    }

    def get_expected_info(label):
        """
        Returns (typeHint, expectedElemSize) from dictionary, or a fallback.
        Used only for debug warnings; doesn't kill program if mismatch.
        """
        return vector_type_expectations.get(label, vector_type_expectations["(vector)"])

    def read_vector(label="(vector)"):
        """
        Reads:
          1) size_val   (number of elements)
          2) elem_val   (bytes per element)
          3) data_buf
        Usually matches the usage 'packer.pack_vector(...)' in the C++ code.
        """
        size_val = read_int(label + ".size")
        elem_val = read_int(label + ".elemSize")
        data_buf = read_bytes(label + ".data")

        (type_hint, expected_size) = get_expected_info(label)
        if expected_size is not None and elem_val != expected_size:
            logger.warning(f"For '{label}', expected elemSize={expected_size} but got {elem_val}.")
        return (size_val, elem_val, data_buf)

    # --------------------------------------------------------------------------------
    # decode_integerset: used for reading TSIntegerSet bits (like rotationMatters, etc.)
    # --------------------------------------------------------------------------------
    def decode_integerset():
        val = next_obj()
        if isinstance(val, bytes):
            return val
        elif isinstance(val, list):
            return val
        else:
            logger.error(f"decode_integerset: got {type(val)} => {val}, not bytes/list. Exiting.")
            sys.exit(1)

    # --------------------------------------------------------------------------------
    # decode_mesh_flags: for reading top bits from tsMesh.h
    # --------------------------------------------------------------------------------
    def decode_mesh_flags(flags: int) -> str:
        """
        Billboard       = 1<<31,
        HasDetailTexture= 1<<30,
        BillboardZAxis  = 1<<29,
        UseEncodedNormals=1<<28,
        ...
        """
        results = []
        if flags & 0x80000000:
            results.append("Billboard")
        if flags & 0x40000000:
            results.append("HasDetailTexture")
        if flags & 0x20000000:
            results.append("BillboardZAxis")
        if flags & 0x10000000:
            results.append("UseEncodedNormals")
        return ", ".join(results) if results else "none"

    # --------------------------------------------------------------------------------
    # 2) Reading data in the order from _write_v30
    # --------------------------------------------------------------------------------

    # a) "Welcome..." string
    welcome_str = read_str("welcome")
    logger.info(f"Welcome string: {welcome_str}")

    # b) number of "objectNames"
    num_objects = read_uint32("numObjects")
    logger.info(f"num_objects={num_objects}")
    object_names = []
    for _ in range(num_objects):
        name = read_str("objectName")
        object_names.append(name)
    logger.debug(f"Object names read: {object_names}")

    # c) shape-level floats, bounding data
    smallest_visible_size = read_float("mSmallestVisibleSize")
    smallest_visible_dl   = read_int("mSmallestVisibleDL")
    radius                = read_float("radius")
    tube_radius           = read_float("tubeRadius")
    center                = read_point3f("center")
    bounds                = read_box6f("bounds")

    logger.info(f"smallest_visible_size={smallest_visible_size}, smallest_visible_dl={smallest_visible_dl}")
    logger.info(f"radius={radius}, tube_radius={tube_radius}")
    logger.info(f"center={center}, bounds={bounds}")

    # d) parse vectors in the order
    nodes_vec = read_vector("nodes")
    logger.debug(f"nodes_vec => size={nodes_vec[0]}, elemSize={nodes_vec[1]}, data_len={len(nodes_vec[2])}")

    # parse Node array -> 5 i32
    num_nodes = nodes_vec[0]
    node_elem_size = nodes_vec[1]
    node_data_buf = nodes_vec[2]
    if node_elem_size != 20:
        logger.warning(f"For 'nodes', expected 20 bytes each, got {node_elem_size}!")

    node_list = []
    for i in range(num_nodes):
        offset = i * 20
        chunk = node_data_buf[offset : offset + 20]
        (nNameIdx, nParentIdx, firstObj, firstChild, nextSibling) = struct.unpack("<5i", chunk)
        node_list.append({
            "nameIndex":   nNameIdx,
            "parentIndex": nParentIdx,
            "firstObject": firstObj,
            "firstChild":  firstChild,
            "nextSibling": nextSibling
        })

    logger.debug("Parsed nodes:")
    for i, nd in enumerate(node_list[:10]):
        logger.debug(f"  Node[{i}] = {nd}")
    if len(node_list) > 10:
        logger.debug(f"  (Truncated {len(node_list) - 10} more nodes...)")

    # parse objects -> each is 6 i32
    objects_vec = read_vector("objects")
    logger.debug(f"objects_vec => size={objects_vec[0]}, elemSize={objects_vec[1]}, data_len={len(objects_vec[2])}")

    (num_objects_vec, object_elem_size, object_data_buf) = objects_vec
    object_list = []
    if object_elem_size == 24:
        for i in range(num_objects_vec):
            offset = i * 24
            chunk = object_data_buf[offset : offset + 24]
            (oNameIdx, oNumMeshes, oStartMesh, oNodeIdx, oNextSibling, oFirstDecal) = struct.unpack("<6i", chunk)
            object_list.append({
                "nameIndex":      oNameIdx,
                "numMeshes":      oNumMeshes,
                "startMeshIndex": oStartMesh,
                "nodeIndex":      oNodeIdx,
                "nextSibling":    oNextSibling,
                "firstDecal":     oFirstDecal  # DEPRECATED
            })
        logger.debug("Parsed objects:")
        for i, obj in enumerate(object_list[:10]):
            logger.debug(f"  Object[{i}] = {obj}")
        if len(object_list) > 10:
            logger.debug(f"  (Truncated {len(object_list) - 10} more objects...)")

    subshapefn_vec = read_vector("subShapeFirstNode")
    logger.debug(f"subShapeFirstNode => size={subshapefn_vec[0]}, elemSize={subshapefn_vec[1]}, data_len={len(subshapefn_vec[2])}")

    subshapefobj_vec = read_vector("subShapeFirstObject")
    logger.debug(f"subShapeFirstObject => size={subshapefobj_vec[0]}, elemSize={subshapefobj_vec[1]}, data_len={len(subshapefobj_vec[2])}")

    subshapenumN_vec = read_vector("subShapeNumNodes")
    logger.debug(f"subShapeNumNodes => size={subshapenumN_vec[0]}, elemSize={subshapenumN_vec[1]}, data_len={len(subshapenumN_vec[2])}")

    # parse subShapeNumNodes -> array of i32
    (count_subshapeN, elemsize_subshapeN, data_subshapeN) = subshapenumN_vec
    subshapenumN_list = []
    for i in range(count_subshapeN):
        offset = i * 4
        chunk = data_subshapeN[offset : offset + 4]
        (val_i,) = struct.unpack("<i", chunk)
        subshapenumN_list.append(val_i)
    logger.debug(f"subshapenumN_list={subshapenumN_list}")

    total_subshape_nodes = sum(subshapenumN_list)
    if total_subshape_nodes != len(node_list):
        logger.warning(f"Mismatch: subshapes say {total_subshape_nodes} nodes, but read {len(node_list)} total.")

    subshapenumO_vec = read_vector("subShapeNumObjects")
    logger.debug(f"subShapeNumObjects => size={subshapenumO_vec[0]}, elemSize={subshapenumO_vec[1]}, data_len={len(subshapenumO_vec[2])}")

    # Default rotations, translations, etc...
    defrots_vec   = read_vector("defaultRotations")
    deftrans_vec  = read_vector("defaultTranslations")
    noderots_vec  = read_vector("nodeRotations")
    nodetrans_vec = read_vector("nodeTranslations")
    nodeUnif_vec  = read_vector("nodeUniformScales")
    nodeAlign_vec = read_vector("nodeAlignedScales")
    nodeArbFactors_vec = read_vector("nodeArbitraryScaleFactors")
    nodeArbRots_vec    = read_vector("nodeArbitraryScaleRots")
    groundTrans_vec    = read_vector("groundTranslations")
    groundRots_vec     = read_vector("groundRotations")

    logger.debug(f"defaultRotations => size={defrots_vec[0]}, elemSize={defrots_vec[1]}, data_len={len(defrots_vec[2])}")
    logger.debug(f"defaultTranslations => size={deftrans_vec[0]}, elemSize={deftrans_vec[1]}, data_len={len(deftrans_vec[2])}")
    logger.debug(f"nodeRotations => size={noderots_vec[0]}, elemSize={noderots_vec[1]}, data_len={len(noderots_vec[2])}")
    logger.debug(f"nodeTranslations => size={nodetrans_vec[0]}, elemSize={nodetrans_vec[1]}, data_len={len(nodetrans_vec[2])}")
    logger.debug(f"nodeUniformScales => size={nodeUnif_vec[0]}, elemSize={nodeUnif_vec[1]}, data_len={len(nodeUnif_vec[2])}")
    logger.debug(f"nodeAlignedScales => size={nodeAlign_vec[0]}, elemSize={nodeAlign_vec[1]}, data_len={len(nodeAlign_vec[2])}")
    logger.debug(f"nodeArbitraryScaleFactors => size={nodeArbFactors_vec[0]}, elemSize={nodeArbFactors_vec[1]}, data_len={len(nodeArbFactors_vec[2])}")
    logger.debug(f"nodeArbitraryScaleRots => size={nodeArbRots_vec[0]}, elemSize={nodeArbRots_vec[1]}, data_len={len(nodeArbRots_vec[2])}")
    logger.debug(f"groundTranslations => size={groundTrans_vec[0]}, elemSize={groundTrans_vec[1]}, data_len={len(groundTrans_vec[2])}")
    logger.debug(f"groundRotations => size={groundRots_vec[0]}, elemSize={groundRots_vec[1]}, data_len={len(groundRots_vec[2])}")

    # objectStates -> e.g. TSShape::ObjectState (12 bytes each)
    objectStates_vec = read_vector("objectStates")
    logger.debug(f"objectStates => size={objectStates_vec[0]}, elemSize={objectStates_vec[1]}, data_len={len(objectStates_vec[2])}")
    (num_objstates, objstate_elem_size, objstate_data) = objectStates_vec
    objstate_list = []
    if objstate_elem_size == 12:
        for i in range(num_objstates):
            offset = i * 12
            chunk = objstate_data[offset : offset + 12]
            (vis_f, frameIndex, matFrameIndex) = struct.unpack("<fii", chunk)
            objstate_list.append({
                "vis":           vis_f,
                "frameIndex":    frameIndex,
                "matFrameIndex": matFrameIndex
            })
        logger.debug("Parsed objectStates:")
        for i, st in enumerate(objstate_list[:10]):
            logger.debug(f"  ObjectState[{i}] = {st}")
        if len(objstate_list) > 10:
            logger.debug(f"  (Truncated {len(objstate_list) - 10} more objectStates...)")

    # triggers -> we consider them 1 u32 + 1 float => 8 bytes each
    triggers_vec = read_vector("triggers")
    logger.debug(f"triggers => size={triggers_vec[0]}, elemSize={triggers_vec[1]}, data_len={len(triggers_vec[2])}")
    (num_triggers, trigger_elem_size, trigger_data) = triggers_vec
    if trigger_elem_size == 8:
        trigger_list = []
        for i in range(num_triggers):
            offset = i * 8
            chunk = trigger_data[offset : offset + 8]
            (state_u32, pos_f) = struct.unpack("<If", chunk)
            trigger_list.append({
                "state": state_u32,
                "pos":   pos_f
            })
        logger.debug("Parsed triggers:")
        for i, t in enumerate(trigger_list[:10]):
            logger.debug(f"  Trigger[{i}] = {t}")
        if len(trigger_list) > 10:
            logger.debug(f"  (Truncated {len(trigger_list) - 10} more triggers...)")

    # details -> 13 fields => 52 bytes
    details_vec = read_vector("details")
    logger.debug(f"details => size={details_vec[0]}, elemSize={details_vec[1]}, data_len={len(details_vec[2])}")
    (num_details, detail_elem_size, detail_data) = details_vec

    detail_list = []
    if detail_elem_size == 52:
        for i in range(num_details):
            offset = i * 52
            chunk = detail_data[offset : offset + 52]
            (nmIdx, subShapeNum, objDetNum, sizeF,
             avgErr, maxErr, polyCount,
             bbDim, bbDetLvl, bbEqSteps,
             bbPolSteps, bbPolAngle, bbIncludePoles) = struct.unpack("<3i3f i 2i 2I f I", chunk)
            detail_list.append({
                "nameIndex":       nmIdx,
                "subShapeNum":     subShapeNum,
                "objectDetailNum": objDetNum,
                "size":            sizeF,
                "averageError":    avgErr,
                "maxError":        maxErr,
                "polyCount":       polyCount,
                "bbDimension":     bbDim,
                "bbDetailLevel":   bbDetLvl,
                "bbEquatorSteps":  bbEqSteps,
                "bbPolarSteps":    bbPolSteps,
                "bbPolarAngle":    bbPolAngle,
                "bbIncludePoles":  bbIncludePoles
            })
        logger.debug("Parsed details:")
        for i, dt in enumerate(detail_list[:10]):
            logger.debug(f"  Detail[{i}] = {dt}")
        if len(detail_list) > 10:
            logger.debug(f"  (Truncated {len(detail_list) - 10} more details...)")

    # e) read # of 'names', then the names themselves
    num_names = read_uint32("names.size")
    all_names = []
    for _ in range(num_names):
        s = read_str("name")
        all_names.append(s)
    logger.debug(f"num_names={num_names}, e.g. first few={all_names[:5]}")

    # f) read total # of meshes
    total_meshes = read_uint32("meshes.size")
    logger.debug(f"total_meshes={total_meshes}")

    # Now that total_meshes is known, check consistency with objects
    if 'object_list' in locals():
        for i, obj in enumerate(object_list):
            startMeshIndex = obj["startMeshIndex"]
            numMeshes      = obj["numMeshes"]
            if startMeshIndex < 0 or (startMeshIndex + numMeshes) > total_meshes:
                logger.warning(
                    f"Object {i} has invalid mesh range: "
                    f"start={startMeshIndex} numMeshes={numMeshes}, "
                    f"but total_meshes={total_meshes}."
                )
    else:
        logger.debug("No object_list found to validate mesh ranges.")

    # A dictionary mapping mesh type -> name
    mesh_type_map = {
        0: "StandardMesh",
        1: "SkinMesh",
        2: "DecalMesh",
        3: "SortedMesh",
        4: "NullMesh"
    }

    # One for-loop per mesh. We do NOT attempt a subShape->object->mesh nesting here;
    # just read them in a flat manner because that's how it's written.
    for mesh_i in range(total_meshes):
        mesh_type_val = read_uint32("mesh_type")
        mesh_type_name = mesh_type_map.get(mesh_type_val, f"Unknown({mesh_type_val})")
        logger.info(f"Mesh {mesh_i}: type={mesh_type_val} => {mesh_type_name}")

        mesh_numFrames    = read_int("mesh_numFrames")
        mesh_numMatFrames = read_int("mesh_numMatFrames")
        mesh_parentMesh   = read_int("mesh_parentMesh")
        mesh_bounds       = read_box6f("mesh_bounds")
        mesh_center       = read_point3f("mesh_center")
        mesh_radius       = read_float("mesh_radius")

        # If it's a NullMesh, we skip the debug prints but still read the data
        if mesh_type_val == 4:
            logger.debug("    [Ignoring null mesh data]")
            verts_vec       = read_vector("mesh.verts")
            tverts_vec      = read_vector("mesh.tverts")
            tverts2_vec     = read_vector("mesh.tverts2")
            colors_vec      = read_vector("mesh.colors")
            norms_vec       = read_vector("mesh.norms")
            encnorms_vec    = read_vector("mesh.encodedNorms")
            prims_vec       = read_vector("mesh.primitives")
            indices_vec     = read_vector("mesh.indices")
            tangents_vec    = read_vector("mesh.tangents")
            mesh_vertsPerFrame = read_int("mesh.vertsPerFrame")
            mesh_flags         = read_uint32("mesh.flags")
            continue

        logger.debug(f"    frames={mesh_numFrames}, matframes={mesh_numMatFrames}, radius={mesh_radius}")

        verts_vec       = read_vector("mesh.verts")
        logger.debug(f"    mesh.verts => size={verts_vec[0]}, elemSize={verts_vec[1]}, data_len={len(verts_vec[2])}")

        tverts_vec      = read_vector("mesh.tverts")
        logger.debug(f"    mesh.tverts => size={tverts_vec[0]}, elemSize={tverts_vec[1]}, data_len={len(tverts_vec[2])}")

        tverts2_vec     = read_vector("mesh.tverts2")
        logger.debug(f"    mesh.tverts2 => size={tverts2_vec[0]}, elemSize={tverts2_vec[1]}, data_len={len(tverts2_vec[2])}")

        colors_vec      = read_vector("mesh.colors")
        logger.debug(f"    mesh.colors => size={colors_vec[0]}, elemSize={colors_vec[1]}, data_len={len(colors_vec[2])}")
        # Possibly parse them as 4 bytes => RGBA
        (color_count, color_elem_size, color_data_buf) = colors_vec
        if color_elem_size == 4:
            color_list = []
            for i in range(color_count):
                offset = i * 4
                chunk = color_data_buf[offset : offset + 4]
                (r, g, b, a) = struct.unpack("<4B", chunk)
                color_list.append({"r": r, "g": g, "b": b, "a": a})
            # Show only first 2 for brevity
            for i, c in enumerate(color_list[:2]):
                hex_rgba = f"#{c['r']:02X}{c['g']:02X}{c['b']:02X}{c['a']:02X}"
                logger.debug(f"     colors[{i}] = {c}  => {hex_rgba}")
            if color_count > 2:
                logger.debug(f"     (Truncated {color_count - 2} more colors...)")

        norms_vec       = read_vector("mesh.norms")
        logger.debug(f"    mesh.norms => size={norms_vec[0]}, elemSize={norms_vec[1]}, data_len={len(norms_vec[2])}")

        encnorms_vec    = read_vector("mesh.encodedNorms")
        logger.debug(f"    mesh.encodedNorms => size={encnorms_vec[0]}, elemSize={encnorms_vec[1]}, data_len={len(encnorms_vec[2])}")

        prims_vec       = read_vector("mesh.primitives")
        logger.debug(f"    mesh.primitives => size={prims_vec[0]}, elemSize={prims_vec[1]}, data_len={len(prims_vec[2])}")
        (prim_count, prim_elem_size, prim_data_buf) = prims_vec
        if prim_elem_size == 12:
            # parse TSDrawPrimitive => 3 i32
            primitive_list = []
            for i in range(prim_count):
                offset = i * 12
                chunk = prim_data_buf[offset : offset + 12]
                (start_i, numElems_i, matIndex_i) = struct.unpack("<3i", chunk)

                # decode matIndex bits
                TYPE_MASK     = 0xC0000000
                INDEXED       = 0x20000000
                NO_MATERIAL   = 0x10000000
                MATERIAL_MASK = 0x0FFFFFFF

                actual_mat_idx = matIndex_i & MATERIAL_MASK
                raw_type_bits  = matIndex_i & TYPE_MASK
                is_indexed     = bool(matIndex_i & INDEXED)
                has_no_mat     = bool(matIndex_i & NO_MATERIAL)

                if raw_type_bits == 0x00000000:
                    draw_type_str = "Triangles"
                elif raw_type_bits == 0x40000000:
                    draw_type_str = "Strip"
                elif raw_type_bits == 0x80000000:
                    draw_type_str = "Fan"
                else:
                    draw_type_str = f"Unknown(0x{raw_type_bits:X})"

                primitive_list.append({
                    "start":       start_i,
                    "numElements": numElems_i,
                    "matIndex":    matIndex_i,
                    "decoded": {
                        "materialIndex": actual_mat_idx,
                        "type":          draw_type_str,
                        "isIndexed":     is_indexed,
                        "noMaterial":    has_no_mat
                    }
                })

            for i, p in enumerate(primitive_list[:10]):
                base_info = f"     Primitive[{i}] => start={p['start']} numElem={p['numElements']}"
                decode_info= p["decoded"]
                logger.debug(f"{base_info}, matIndex={p['matIndex']} => "
                             f"{{ materialIndex={decode_info['materialIndex']}, "
                             f"type={decode_info['type']}, isIndexed={decode_info['isIndexed']}, "
                             f"noMaterial={decode_info['noMaterial']} }}")
            if len(primitive_list) > 10:
                logger.debug(f"     (Truncated {len(primitive_list) - 10} more primitives...)")

        indices_vec = read_vector("mesh.indices")
        logger.debug(f"    mesh.indices => size={indices_vec[0]}, elemSize={indices_vec[1]}, data_len={len(indices_vec[2])}")

        tangents_vec = read_vector("mesh.tangents")
        logger.debug(f"    mesh.tangents => size={tangents_vec[0]}, elemSize={tangents_vec[1]}, data_len={len(tangents_vec[2])}")

        mesh_vertsPerFrame = read_int("mesh.vertsPerFrame")
        logger.debug(f"    mesh.vertsPerFrame={mesh_vertsPerFrame}")

        mesh_flags = read_uint32("mesh.flags")
        decoded_flags = decode_mesh_flags(mesh_flags)
        logger.debug(f"    mesh.flags=0x{mesh_flags:X} => {decoded_flags}")

    # g) sequences
    tot_sequences = read_uint32("sequences.size")
    sequences_info = []
    for seq_i in range(tot_sequences):
        seq_nameIndex      = read_int("seq.nameIndex")
        seq_flags          = read_int("seq.flags")
        seq_numKeyframes   = read_int("seq.numKeyframes")
        seq_duration       = read_float("seq.duration")
        seq_priority       = read_int("seq.priority")
        seq_firstGroundFrame = read_int("seq.firstGroundFrame")
        seq_numGroundFrames  = read_int("seq.numGroundFrames")
        seq_baseRotation     = read_int("seq.baseRotation")
        seq_baseTranslation  = read_int("seq.baseTranslation")
        seq_baseScale        = read_int("seq.baseScale")
        seq_baseObjectState  = read_int("seq.baseObjectState")
        seq_baseDecalState   = read_int("seq.baseDecalState")
        seq_firstTrigger     = read_int("seq.firstTrigger")
        seq_numTriggers      = read_int("seq.numTriggers")
        seq_toolBegin        = read_float("seq.toolBegin")

        # reading TSIntegerSets
        seq_rotationMatters    = decode_integerset()
        seq_translationMatters = decode_integerset()
        seq_scaleMatters       = decode_integerset()
        seq_visMatters         = decode_integerset()
        seq_frameMatters       = decode_integerset()
        seq_matFrameMatters    = decode_integerset()

        sequences_info.append({
            "nameIndex": seq_nameIndex,
            "flags": seq_flags,
            "numKeyframes": seq_numKeyframes,
            "duration": seq_duration,
            "priority": seq_priority,
            "firstGroundFrame": seq_firstGroundFrame,
            "numGroundFrames": seq_numGroundFrames
            # more fields if needed...
        })
    logger.info(f"[INFO] read {tot_sequences} sequences")

    # h) material list (if present)
    leftover = True
    mat_list_size = None
    try:
        mat_list_size = read_uint32("materialList.size")
    except:
        leftover = False

    if leftover and mat_list_size is not None:
        materials_data = []
        for _ in range(mat_list_size):
            mat_name = read_str("material.name")
            mat_flags = read_uint32("material.flags")
            mat_reflect = read_uint32("material.reflectMap")
            mat_bump    = read_uint32("material.bumpMap")
            mat_detail  = read_uint32("material.detailMap")
            mat_detailScale = read_float("material.detailScale")
            mat_reflectionAmt = read_float("material.reflectionAmount")
            materials_data.append((
                mat_name,
                mat_flags,
                mat_reflect,
                mat_bump,
                mat_detail,
                mat_detailScale,
                mat_reflectionAmt
            ))

        logger.info(f"[INFO] read {mat_list_size} materials from materialList.")

        # Log each material's name and decoded flags
        for i, mat in enumerate(materials_data):
            (mname, mflags, mreflect, mbump, mdetail,
             mdetailScale, mrefAmt) = mat
            decoded_flags = decode_material_flags(mflags)
            logger.info(
                f"  Material[{i}]: name='{mname}', "
                f"flags=0x{mflags:X} => {decoded_flags}, "
                f"reflectMap={mreflect}, bumpMap={mbump}, "
                f"detailMap={mdetail}, detailScale={mdetailScale}, "
                f"reflectionAmt={mrefAmt}"
            )
    else:
        logger.info("[INFO] No material list found or no leftover data to parse for materials.")

    logger.info("[DONE] Successfully read shape data in the order of _write_v30.")

# --------------------------------------------------------------------------------
# Main script entry point
# --------------------------------------------------------------------------------

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Strict v30 TSShape parser matching the _write_v30() order.")
    parser.add_argument("filepath", help="Path to the .v30 shape file")
    args = parser.parse_args()
    parse_v30_file(args.filepath)
